// A simple/fast(?) hash table.

#ifndef ED_HT_INCLUDED
#define ED_HT_INCLUDED

#include <stdlib.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>

#include "string.h"

// see <https://en.wikipedia.org/wiki/Fowler–Noll–Vo_hash_function>
#define FNV_OFFSET 14695981039346656037ULL
#define FNV_PRIME 1099511628211ULL

typedef struct {
    string key;
} ed_ht_slot;

typedef struct {
    ed_ht_slot *key_slots;
    void *value_slots;
    size_t value_size;
    size_t capacity;
} ed_ht;

typedef uint64_t ht_hash_t;

#ifdef ED_HT_IMPLEMENTATION

static ht_hash_t ht_hash(string key, uint64_t mod) {
    ht_hash_t hash = FNV_OFFSET;

    for (size_t i=0; i<key.len; ++i) {
        hash *= FNV_PRIME;
        hash ^= key.data[i];
    }

    return hash % mod;
}

ed_ht ht_create(size_t max_entries, size_t value_size) {
    ed_ht_slot *key_slots = (ed_ht_slot *)malloc(max_entries * sizeof(ed_ht_slot));
    void *value_slots = malloc(max_entries * value_size);

    memset(key_slots, 0, max_entries * sizeof(ed_ht_slot));
    memset(value_slots, 0, max_entries * value_size);

    return (ed_ht) {
        .key_slots = key_slots,
        .value_slots = value_slots,
        .value_size = value_size,
        .capacity = max_entries,
    };
}

bool ht_set(ed_ht *ht, string key, void *value) {
    ht_hash_t hash = ht_hash(key, ht->capacity);

    for (size_t i=hash; i<ht->capacity; ++i) {
        if (ht->key_slots[i].key.data == NULL || string_eq(ht->key_slots[i].key, key)) {
            ht->key_slots[i].key = key;

            memcpy(ht->value_slots+i*ht->value_size, value, ht->value_size);
            return true;
        }
    }

    return false;
}

void *ht_get_slot(ed_ht *ht, size_t slot) {
    void *value = ht->value_slots+slot*ht->value_size;

    if (slot >= ht->capacity || !value)
        return NULL;

    return value;
}

void *ht_get(ed_ht *ht, string key) {
    ht_hash_t hash = ht_hash(key, ht->capacity);

    for (size_t i=hash; i<ht->capacity; ++i) {
        if (ht->key_slots[i].key.data == NULL) {
            return NULL;
        }

        if (string_eq(ht->key_slots[i].key, key)) {
            return ht_get_slot(ht, i);
        }
    }

    return NULL;
}

void ht_destroy(ed_ht *ht) {
    // TODO: destroy the hash table
    free(ht->key_slots);
    free(ht->value_slots);

    ht->key_slots = NULL;
    ht->value_slots = NULL;
}

#endif
#endif
